@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 import sys
 import os
-import json
+import ujson
 import fido
 import time
 import gzip
@@ -48,12 +48,13 @@ from pint import UnitRegistry
 from pint.unit import UnitDefinition
 from pint.converters import ScaleConverter
 from pint.errors import DimensionalityError
-from datetime import datetime
 
 RETRIES = 3
-MAX_WORKERS = 10
+MAX_WORKERS = 8
 MAX_ELEMS = 10
-MAX_BYTES = 1200 * 1024
+MEGABYTES = 1024 * 1024
+MAX_BYTES = 1.2 * MEGABYTES
+MAX_PAYLOAD = 15 * MEGABYTES
 DEFAULT_HOST = "contribs-api.materialsproject.org"
 BULMA = "is-narrow is-fullwidth has-background-light"
 PROVIDERS = {"github", "google", "facebook", "microsoft", "amazon"}
@@ -98,7 +99,7 @@ ureg.define("electron_mass = 9.1093837015e-31 kg = mₑ = m_e")
 
 
 def get_md5(d):
-    s = json.dumps(d, sort_keys=True).encode("utf-8")
+    s = ujson.dumps(d, sort_keys=True).encode("utf-8")
     return md5(s).hexdigest()
 
 
@@ -145,7 +146,7 @@ def grouper(n, iterable):
         yield chunk
 
 
-def get_session():
+def get_session(max_workers=MAX_WORKERS):
     # TODO add Bad Gateway 502?
     adapter_kwargs = dict(max_retries=Retry(
         total=RETRIES,
@@ -154,7 +155,7 @@ def get_session():
         respect_retry_after_header=True,
         status_forcelist=[429],  # rate limit
     ))
-    s = FuturesSession(max_workers=MAX_WORKERS, adapter_kwargs=adapter_kwargs)
+    s = FuturesSession(max_workers=max_workers, adapter_kwargs=adapter_kwargs)
     s.hooks['response'].append(_response_hook)
     return s
 
@@ -177,8 +178,10 @@ def _response_hook(resp, *args, **kwargs):
 
     elif content_type == "application/gzip":
         resp.result = resp.content
+        resp.count = 1
     else:
         print("ERROR", resp.status_code)
+        resp.count = 0
 
 
 class FidoClientGlobalHeaders(FidoClient):
@@ -261,6 +264,14 @@ class Attachment(dict):
         """Decode base64-encoded content of attachment"""
         return b64decode(self["content"], validate=True)
 
+    def unpack(self) -> str:
+        unpacked = self.decode()
+
+        if self["mime"] == "application/gzip":
+            unpacked = gzip.decompress(unpacked).decode("utf-8")
+
+        return unpacked
+
     def write(self, outdir: Union[str, Path] = None) -> Path:
         """Write attachment to file using its name
 
@@ -310,7 +321,7 @@ class Attachment(dict):
             data (list,dict): JSON-serializable data to go into the attachment
         """
         filename = name + ".json.gz"
-        data_json = json.dumps(data, indent=4).encode("utf-8")
+        data_json = ujson.dumps(data, indent=4).encode("utf-8")
         content = gzip.compress(data_json)
         size = len(content)
 
@@ -330,14 +341,14 @@ def load_client(apikey=None, headers=None, host=None):
     )
 
 
-def _run_futures(futures, total: int = 0, timeout: int = -1):
+def _run_futures(futures, total: int = 0, timeout: int = -1, desc=None):
     """helper to run futures/requests"""
     start = time.perf_counter()
     total_set = total > 0
     total = total if total_set else len(futures)
     responses = {}
 
-    with tqdm(leave=False, total=total) as pbar:
+    with tqdm(total=total, desc=desc) as pbar:
         for future in as_completed(futures):
             if not future.cancelled():
                 response = future.result()
@@ -346,10 +357,11 @@ def _run_futures(futures, total: int = 0, timeout: int = -1):
 
                 if hasattr(future, "track_id"):
                     tid = future.track_id
+                    responses[tid] = {}
                     if hasattr(response, "result"):
-                        responses[tid] = response.result
-                    elif hasattr(response, "count"):
-                        responses[tid] = {"count": response.count}
+                        responses[tid]["result"] = response.result
+                    if hasattr(response, "count"):
+                        responses[tid]["count"] = response.count
 
                 elapsed = time.perf_counter() - start
                 timed_out = timeout > 0 and elapsed > timeout
@@ -583,7 +595,7 @@ class Client(SwaggerClient):
         kwargs = dict(headers=self.headers, params=params)
 
         if method == "put" and data:
-            kwargs["data"] = json.dumps(data).encode("utf-8")
+            kwargs["data"] = ujson.dumps(data).encode("utf-8")
 
         future = getattr(self.session, method)(
             f"{self.url}/{rel_url}/", **kwargs
@@ -905,15 +917,15 @@ class Client(SwaggerClient):
         query = query or {}
         skip_keys = {"per_page", "_fields", "format"}
         query = {k: v for k, v in query.items() if k not in skip_keys}
-        query["_fields"] = ["id"]
+        query["_fields"] = []  # only need totals -> explicitly request no fields
         queries = self._split_query(query, resource=resource, op=op)  # don't paginate
         result = {"total_count": 0, "total_pages": 0}
         futures = [self._get_future(i, q, rel_url=resource) for i, q in enumerate(queries)]
-        responses = _run_futures(futures, timeout=timeout)
+        responses = _run_futures(futures, timeout=timeout, desc="Totals")
 
         for resp in responses.values():
             for k in result:
-                result[k] += resp[k]
+                result[k] += resp["result"][k]
 
         return result["total_count"], result["total_pages"]
 
@@ -1030,10 +1042,10 @@ class Client(SwaggerClient):
         _, total_pages = self.get_totals(query=query, timeout=timeout)
         queries = self._split_query(query, op=op, pages=total_pages)
         futures = [self._get_future(i, q) for i, q in enumerate(queries)]
-        responses = _run_futures(futures, timeout=timeout)
+        responses = _run_futures(futures, timeout=timeout, desc="Identifiers")
 
         for resp in responses.values():
-            for contrib in resp["data"]:
+            for contrib in resp["result"]["data"]:
                 project = contrib["project"]
                 data_id_field = data_id_fields.get(project)
 
@@ -1237,7 +1249,8 @@ class Client(SwaggerClient):
         ignore_dupes: bool = False,
         retry: bool = False,
         per_request: int = 100,
-        timeout: int = -1
+        timeout: int = -1,
+        skip_dupe_check: bool = False
     ):
         """Submit a list of contributions
 
@@ -1267,6 +1280,7 @@ class Client(SwaggerClient):
             retry (bool): keep trying until all contributions successfully submitted
             per_request (int): number of contributions to submit per request
             timeout (int): cancel remaining requests if timeout exceeded (in seconds)
+            skip_dupe_check (bool): skip duplicate check for contribution identifiers
         """
         if not contributions or not isinstance(contributions, list):
             print("Please provide list of contributions to submit.")
@@ -1306,17 +1320,16 @@ class Client(SwaggerClient):
                     id2project[cid] = project_name
 
         existing = defaultdict(dict)
+        unique_identifiers = defaultdict(dict)
         project_names = list(project_names)
 
-        if len(collect_ids) != len(contributions):
-            print("get existing contributions ...")
+        if not skip_dupe_check and len(collect_ids) != len(contributions):
             unique_identifiers = self.get_unique_identifiers_flags(projects=project_names)
             existing = defaultdict(dict, self.get_all_ids(
                 dict(project__in=project_names), include=COMPONENTS, timeout=timeout
             ))
 
         # prepare contributions
-        print("prepare contributions ...")
         contribs = defaultdict(list)
         digests = {project_name: defaultdict(set) for project_name in project_names}
         fields = [
@@ -1326,12 +1339,12 @@ class Client(SwaggerClient):
         ]
         fields.remove("needs_build")  # internal field
 
-        for contrib in tqdm(contributions, leave=False):
+        for contrib in tqdm(contributions, desc="Prepare"):
             update = "id" in contrib
             project_name = id2project[contrib["id"]] if update else contrib["project"]
             if (
-                not update and unique_identifiers[project_name]
-                and contrib["identifier"] in existing[project_name].get("identifiers", {})
+                not update and unique_identifiers.get(project_name)
+                and contrib["identifier"] in existing.get(project_name, {}).get("identifiers", {})
             ):
                 continue
 
@@ -1436,7 +1449,7 @@ class Client(SwaggerClient):
 
                     dupe = bool(
                         digest in digests[project_name][component] or
-                        digest in existing[project_name].get(component, {}).get("md5s", set())
+                        digest in existing.get(project_name, {}).get(component, {}).get("md5s", [])
                     )
 
                     if not ignore_dupes and dupe:
@@ -1456,68 +1469,98 @@ class Client(SwaggerClient):
             if self.session and self.session.executor._shutdown:
                 raise ValueError("Session closed. Use `with` statement.")
 
-            print("submit contributions ...")
-            total = 0
+            total, total_processed = 0, 0
+            self.session.close()
+            self.session = get_session(max_workers=2)
 
-            def post_future(chunk):
-                return self.session.post(
+            def post_future(track_id, payload):
+                future = self.session.post(
                     f"{self.url}/contributions/",
                     headers=self.headers,
-                    data=json.dumps(chunk).encode("utf-8"),
+                    data=payload,
                 )
+                setattr(future, "track_id", track_id)
+                return future
 
-            def put_future(cdct):
-                pk = cdct.pop("id")
-                return self.session.put(
+            def put_future(pk, payload):
+                future = self.session.put(
                     f"{self.url}/contributions/{pk}/",
                     headers=self.headers,
-                    data=json.dumps(cdct).encode("utf-8"),
+                    data=payload,
                 )
+                setattr(future, "track_id", pk)
+                return future
 
             for project_name in project_names:
                 ncontribs = len(contribs[project_name])
                 total += ncontribs
-                start = datetime.utcnow()
+                retries = 0
 
                 while contribs[project_name]:
                     futures = []
-                    for chunk in grouper(per_page, contribs[project_name]):
+                    for idx, chunk in enumerate(grouper(per_page, contribs[project_name])):
                         post_chunk = []
                         for c in chunk:
                             if "id" in c:
-                                futures.append(put_future(c))
+                                pk = c.pop("id")
+                                payload = ujson.dumps(c).encode("utf-8")
+                                if len(payload) < MAX_PAYLOAD:
+                                    futures.append(put_future(pk, payload))
+                                else:
+                                    print(
+                                        f"SKIPPED update of {project_name}/{pk}: too large."
+                                    )
                             else:
                                 post_chunk.append(c)
 
                         if post_chunk:
-                            futures.append(post_future(post_chunk))
+                            payload = ujson.dumps(post_chunk).encode("utf-8")
+                            if len(payload) < MAX_PAYLOAD:
+                                futures.append(post_future(idx, payload))
+                            else:
+                                print(
+                                    f"SKIPPED {project_name}/{idx}: too large, reduce per_request"
+                                )
 
-                    _run_futures(futures, total=ncontribs, timeout=timeout)
+                    if not futures:
+                        break  # nothing to do
 
-                    if retry and unique_identifiers[project_name]:
+                    responses = _run_futures(
+                        futures, total=ncontribs, timeout=timeout, desc="Submit"
+                    )
+                    processed = sum(r["count"] for r in responses.values())
+                    total_processed += processed
+
+                    if processed != ncontribs and retry and retries < RETRIES and \
+                            unique_identifiers.get(project_name):
                         existing[project_name] = self.get_all_ids(
                             dict(project=project_name), include=COMPONENTS, timeout=timeout
                         ).get(project_name, {"identifiers": set()})
                         unique_identifiers[project_name] = self.projects.get_entry(
                             pk=project_name, _fields=["unique_identifiers"]
                         ).result()["unique_identifiers"]
+                        existing_ids = existing.get(project_name, {}).get("identifiers", [])
                         contribs[project_name] = [
                             c for c in contribs[project_name]
-                            if c["identifier"] not in existing[project_name]["identifiers"]
+                            if c["identifier"] not in existing_ids
                         ]
+                        retries += 1
                     else:
                         contribs[project_name] = []  # abort retrying
-                        if retry and not unique_identifiers[project_name]:
-                            print("Please resubmit failed contributions manually.")
+                        if processed != ncontribs and retry:
+                            if retries >= RETRIES:
+                                print(f"{project_name}: Tried {RETRIES} times - abort.")
+                            elif not unique_identifiers.get(project_name):
+                                print(
+                                    f"{project_name}: resubmit failed contributions manually"
+                                )
 
-            end = datetime.utcnow()
-            updated_total, _ = self.get_totals(query=dict(
-                project__in=project_names, last_modified__after=start, last_modified__before=end
-            ), timeout=timeout)
             toc = time.perf_counter()
             dt = (toc - tic) / 60
             self._load()
-            print(f"It took {dt:.1f}min to submit {updated_total} contributions.")
+            self.session.close()
+            self.session = get_session()
+            print(f"It took {dt:.1f}min to submit {total_processed}/{total} contributions.")
         else:
             print("Nothing to submit.")
 
@@ -1730,7 +1773,7 @@ class Client(SwaggerClient):
             responses = _run_futures(futures, timeout=timeout)
 
             for path, resp in responses.items():
-                path.write_bytes(resp)
+                path.write_bytes(resp["result"])
                 paths.append(path)
 
         return paths
